package ecaml

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Ecaml.BookmarkSource

(Info-goto-node "(emacs) Bookmarks")

Sourcetype t
Sourceval sexp_of_t : t -> Ppx_sexp_conv_lib.Sexp.t
include Core_kernel.Comparator.S with type t := t
Sourcetype comparator_witness
include Core_kernel.Stringable.S with type t := t
Sourceval of_string : string -> t
Sourceval to_string : t -> string
include Ecaml_value.Valueable.S with type t := t
Sourceval of_value_exn : Ecaml_value__.Value0.t -> t
Sourceval to_value : t -> Ecaml_value__.Value0.t
Sourcemodule Property : sig ... end
Sourcemodule Record : sig ... end

The data side of bookmark-alist

Sourceval set : t -> Record.t -> no_overwrite:bool -> unit

set m t a ~no_overwrite:true means, let a shadow any previous location bookmarked as t. If the user removes this bookmark, the previous bookmark of this name will be restored.

Sourceval param : (Record.t -> 'a) -> 'a Defun.t
Sourcemodule Make_record_function : sig ... end